package ru.dbotthepony.kstarbound.client.gl.program

import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.shader.GLShader
import ru.dbotthepony.kstarbound.client.gl.shader.GLUniformLocation
import ru.dbotthepony.kstarbound.client.gl.shader.ShaderLinkException
import ru.dbotthepony.kvector.api.IFloatMatrix
import ru.dbotthepony.kvector.api.IStruct3f
import ru.dbotthepony.kvector.api.IStruct4f
import java.util.*
import kotlin.collections.HashSet

open class GLShaderProgram(val state: GLStateTracker, vararg shaders: GLShader) {
	init {
		state.ensureSameThread()
	}

	val pointer = glCreateProgram()
	var linked = false
		private set

	private val attached = HashSet<GLShader>()
	val access = object : Collection<GLShader> by attached {}

	private val locationCache = Object2ObjectOpenHashMap<String, Optional<GLUniformLocation>>()

	/**
	 * Возвращает GLUniformLocation или null, если у данной программы нет такого uniform
	 *
	 * В т.ч. если она была в коде шейдера, но нигде не использовалась (отсечена компилятором)
	 *
	 * Результат поиска кешируется, но для повышения производительности вызывающий код желательно так же
	 * должен кешировать результат (как локальное свойство или переменная, в случае многократного использования)
	 */
	operator fun get(name: String): GLUniformLocation? {
		state.ensureSameThread()
		check(linked) { "Shader program is not linked!" }

		return locationCache.computeIfAbsent(name, Object2ObjectFunction {
			val location = glGetUniformLocation(pointer, name)

			if (location == -1)
				return@Object2ObjectFunction Optional.empty<GLUniformLocation>()

			return@Object2ObjectFunction Optional.of(GLUniformLocation(this, name, location))
		}).orElse(null)
	}

	operator fun set(name: String, value: IStruct4f) = this[name]?.set(value)
	operator fun set(name: String, value: IStruct3f) = this[name]?.set(value)
	operator fun set(name: String, value: Int) = this[name]?.set(value)
	operator fun set(name: String, value: IFloatMatrix<*>) = this[name]?.set(value)

	fun attach(shader: GLShader) {
		state.ensureSameThread()
		check(!linked) { "Program is already linked!" }

		if (!attached.add(shader)) {
			throw IllegalStateException("Already attached! $shader")
		}
		require(!shader.unlinked) { "$shader is already unlinked, and thus can not be used" }

		glAttachShader(pointer, shader.pointer)
	}

	fun link() {
		check(!linked) { "Program is already linked!" }
		state.ensureSameThread()
		glLinkProgram(pointer)

		val success = intArrayOf(0)
		glGetProgramiv(pointer, GL_LINK_STATUS, success)

		if (success[0] == 0) {
			throw ShaderLinkException(glGetShaderInfoLog(pointer))
		}

		glGetError()

		linked = true
	}

	fun use() = state.use(this)

	fun unlinkChildren() {
		attached.forEach(GLShader::unlink)
	}

	init {
		shaders.forEach(this::attach)
	}
}
